/************************************************************************
 *   IRC - Internet Relay Chat, src/m_nick.c
 *   Copyright (C) 2000 Lucas Madar [bahamut team]
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: m_nick.c,v 1.27 2002/04/11 19:43:44 lucas Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "h.h"
#include "throttle.h"

extern int do_user(char *, aClient *, aClient *, char *, char *, char *, 
		   unsigned long, char *);
extern int user_mode_table[];
extern int do_nick_name(char *);
extern int register_user(aClient *, aClient *, char *, char *);
extern char *inet_ntoa(struct in_addr);

/*
 * s_nick 
 * parv[0] = sender prefix 
 * parv[1] = nickname 
 * parv[2] = hopcount when new user; TS when nick change 
 * ---- new user only below ---- 
 * parv[3] = TS
 * parv[4] = umode 
 * parv[5] = username 
 * parv[6] = hostname 
 * parv[7] = server, OR base64 client/serverid pair -- if begins with '!'
 * parv[8] = serviceid
 * parv[9] = ip addr (if sptr ! TOK1)
 * parv[10] = ircname
 */

int s_nick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
#ifndef DONT_CHECK_QLINE_REMOTE
   aConfItem  *aconf;
#endif
   aClient    *acptr, *uplink;
   char        nick[NICKLEN + 2];
   ts_val      newts = 0;
   int         sameuser = 0, samenick = 0, numbersare64 = 0;
   unsigned long flag;
   char *mptr;
        
   /*
    * parc == 4 on a normal server-to-server client nick change
    * parc == 10 on a normal TS style server-to-server NICK introduction
    * parc == 11 on a nickip server-to-server NICK introduction
    */

   if (IsServer(sptr)) /* this is a new remote nick */
   {
     /* check to make sure we've got the right number of params
      * if not, ignore the nick completely.
      */
     	if ((((parc > 4) && (parc < 10)) || (IsNickIP(sptr) && ((parc > 4) && (parc < 11)))))
     	{
        	sendto_realops("IGNORING BAD NICK: %s[%s@%s] on %s (from %s)", parv[1],
                     (parc >= 6) ? parv[5] : "-",
                     (parc >= 7) ? parv[6] : "-",
                     (parc >= 8) ? parv[7] : "-", parv[0]);
        	return 0;

     	}

      	if(*parv[3] == '!')
      	{
         	numbersare64 = 1;
         	newts = base64dec(parv[3] + 1);
      	}
      	else
         	newts = atol(parv[3]); 

   }
   else
      	newts = atol(parv[2]); /* this is a remote user changing nicks */

   strncpyzt(nick, parv[1], NICKLEN + 1);

   /* if do_nick_name() says the nick is invalid or was changed, reject it */

   if (do_nick_name(nick))
   {
      send_me_numeric(sptr, ERR_ERRONEUSNICKNAME, parv[1]);
      ircstp->is_kill++;
      sendto_realops_lev(DEBUG_LEV, "Bad Nick: %s From: %s %s", parv[1], parv[0],
                         get_client_name(cptr, FALSE));
      sendto_one_client_real(cptr, NULL, &me, &CMD_KILL, "%s :%s (%s <- %s[%s])",
                             parv[1], me.name, parv[1], nick, cptr->name);
      if (sptr != cptr) /* bad remote nick change */
      {     
         send_serv_message_butone(cptr, &me, sptr, &CMD_KILL, ":%s (%s <- %s!%s@%s)", 
                                  me.name, get_client_name(cptr, FALSE), parv[0], 
                                  sptr->user ? sptr->username : "*",
                                  sptr->user ? sptr->user->server : cptr->name);
         sptr->flags |= FLAGS_KILLED;
         return exit_client(cptr, sptr, &me, "Bad Nickname");
      }
      return 0;
   }

   do 
   {
      /* alright. servers should ALWAYS have "." in them, which is an invalid nick char.
       * this should be completely unnecessary. If your network is broken,
       * define this. - lucas */
#ifdef BROKEN_SERVER_NAMES
      if ((acptr = find_server(nick, NULL)))
      {
         /*
          * We have a nickname trying to use the same name as a
          * server. Send out a nick collision KILL to remove the
          * nickname. As long as only a KILL is sent out, there is no
          * danger of the server being disconnected. 
          */
         sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)",
                 sptr->name, acptr->from->name, get_client_name(cptr, FALSE));
         ircstp->is_kill++;
         sendto_one_client_real(cptr, sptr, &me, &CMD_KILL, ":%s (%s <- %s)",
                           me.name, acptr->from->name, get_client_name(cptr, FALSE));
         sptr->flags |= FLAGS_KILLED;
         return exit_client(cptr, sptr, &me, "Nick/Server collision");
      }
#endif /* BROKEN_SERVER_NAMES */

      /* if there is no other client on the network with this nickname, whew!
       * break out of the loop! */
      if (!(acptr = find_client(nick, NULL)))
         break;

      /*
       * If acptr == sptr, then we have a client doing a nick change
       * between *equivalent* nicknames as far as server is concerned
       * (user is changing the case of his/her nickname or somesuch)
       */
      if (acptr == sptr)
      {
         if (strcmp(acptr->name, nick) != 0)
           break;
         else /* case of nickname did not change -- silently ignore it */
           return 0;
      }

      /*
       * Note: From this point forward it can be assumed that acptr != sptr
       * (point to different client structures).
       */

      /*
       * If the older one is "non-person", the new entry is just
       * allowed to overwrite it. Just silently drop non-person, and
       * proceed with the nick. 
       */

      if (IsUnknown(acptr)) /* unknowns _always_ are local to me */
      {
         exit_client(NULL, acptr, &me, "Overridden by older remote signon");
         /* we're exiting acptr, so this nick change is fine */
         break;
      }

      if (IsServer(sptr)) /* this is a new nick being introduced ... */
      {
         if (!newts || !acptr->tsinfo || (newts == acptr->tsinfo))
         {
            sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)(both killed)",
                    acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
            ircstp->is_kill++;
            send_me_numeric(acptr, ERR_NICKCOLLISION, acptr->name);
            send_serv_message_butone(NULL, &me, acptr, &CMD_KILL, ":%s (Nick override, same TS (%s <- %s))",
                                     me.name, acptr->from->name, get_client_name(cptr, FALSE));
            acptr->flags |= FLAGS_KILLED;
            return exit_client(cptr, acptr, &me, "Nick override (same TS)");
         }
         else
         {
	    /* ok, who the fuck wrote this?
 	     * I see what it does, but does this appear like massively backwards
	     * code to anyone else? I'm gonna rewrite this soon.. -epi
             *
             * I agree, this is quite confusing. I've commented it all. -lucas
	     */

            /*
             * same user detection
             * If they have the same user@host, we want to kill off the OLDER client to be nice
             * If they are different, kill off the NEWER client because it's trying to collide.
             *
             * They are the same user if... */
            sameuser = (acptr->user) &&                                /* they are actually a 'user' */
                       (mycmp(acptr->user->username, parv[5]) == 0) && /* they have the same user name */
                       (mycmp(acptr->user->host, parv[6]) == 0);       /* they have the same hostname */

            /*
             * If they're the same user, and the introduced has an older timestamp, OR
             * If they're NOT the same user, and the introduced has a newer timestamp
             */
            if ((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo))
            {
               /* 
                * whoa -- this was done very poorly before. We want to remove 
                * the client which is being introduced. Send out a KILL for it
                * to cptr _only_ -- the good client is on or behind us -- and return.
                * We'll call this a "nick override" -- a nick collide
                * will be for something we already know.
                * DESYNCH POSSIBILITY: 
                * If the next server down the line (cptr) doesn't know about this client
                * already, and it's not currently in the sendQ .. something is broke ;)
                */

               /*
                * this seems to generate a problem in resynchs
                * if our NICK collided someone off remotely, this KILL will kill that user
                * off remotely as well, but not locally. oops.
                * this leads me to believe that our desynch "problem" is not really 
                * a problem in code but rather in time synchronization;
                * one server decides to disagree and simply ignores the new nickname (here)
                * while all the other servers have a different record.
                */

               /*
                * I'm almost 100% sure that this is what leads to things like
                *   server1.dal.net: lucas is lucas@syr-1-2-3-4.twcny.rr.com
                *   server2.dal.net: lucas is ~soab@212.65.96.142
                * What should really be done seems to be:
                *  - send a KILL for this nick down to cptr
                *  - send a NICK for this nick down to cptr
                *
                * I'm not going to introduce this in 1.6.0, as it could potentially cause
                * more desynchs. Once 1.6.0 is stable, code to re-send the NICK
                * should be added below and the below should be enabled.
                */
#if 0
               if (sameuser)
               {
                  sendto_ops_lev(SKILL_LEV, "Nick override on %s(%s -> %s)(older killed)",
                                 acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
                  sendto_one(cptr, ":%s KILL %s :%s (Nick override: Older killed (%s -> %s))", 
                             me.name, acptr->name, me.name, acptr->from->name, 
                             get_client_name(cptr, FALSE));
               }
               else
               {
                  sendto_ops_lev(SKILL_LEV, "Nick override on %s(%s -> %s)(newer killed)",
                                 acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
                  sendto_one(cptr, ":%s KILL %s :%s (Nick override: Newer killed (%s -> %s))", 
                             me.name, acptr->name, me.name, acptr->from->name, 
                             get_client_name(cptr, FALSE));
               }

               ircstp->is_kill++;

               /* this client does not yet have a record on this side of the network.
                * just exit silently */
#endif

               return 0;
            }
            else 
            {
               /* 
                * !((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo))
                *
                * If they're the same user, and the introduced has an newer timestamp, OR
                * If they're NOT the same user, and the introduced has a older timestamp
                */

               /* the incoming client is knocking off the client we already have (acptr) */

               send_me_numeric(acptr, ERR_NICKCOLLISION, acptr->name);

               if (sameuser)
               {
                  sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)(older killed)",
                                 acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
                  send_serv_message_butone(cptr, &me, acptr, &CMD_KILL, ":%s (Nick collide: Older killed (%s <- %s))",
                                           me.name, acptr->from->name, get_client_name(cptr, FALSE));
               }
               else
               {
                  sendto_ops_lev(SKILL_LEV, "Nick collision on %s(%s <- %s)(newer killed)",
                                 acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
                  send_serv_message_butone(cptr, &me, acptr, &CMD_KILL, ":%s (Nick collide: Newer killed (%s <- %s))",
                                           me.name, acptr->from->name, get_client_name(cptr, FALSE));
               }

               ircstp->is_kill++;
               acptr->flags |= FLAGS_KILLED;
               if(sameuser)
                  exit_client(cptr, acptr, &me, "Nick collision (old)");
               else
                  exit_client(cptr, acptr, &me, "Nick collision (new)");
               break; /* let this NICK go through now.. */
            }
         }
      }
      /* This is a user changing nicks... */
      else 
      {
         if (!newts || !acptr->tsinfo || (newts == acptr->tsinfo) || !sptr->user)
         {
            sendto_ops_lev(SKILL_LEV, "Nick change collision from %s to %s(%s <- %s)(both killed)",
                           sptr->name, acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
            ircstp->is_kill++;
            send_me_numeric(acptr, ERR_NICKCOLLISION, acptr->name);
            send_serv_message_butone(NULL, &me, sptr, &CMD_KILL, ":%s (Nick collide, same TS (%s(%s) <- %s))",
                                     me.name, acptr->from->name, acptr->name, get_client_name(cptr, FALSE));
            ircstp->is_kill++;
            send_serv_message_butone(NULL, &me, acptr, &CMD_KILL, ":%s (Nick collide, same TS (%s <- %s(%s)))",
                                     me.name, acptr->from->name, get_client_name(cptr, FALSE), sptr->name);

            acptr->flags |= FLAGS_KILLED;
            exit_client(NULL, acptr, &me, "Nick change collision");
            sptr->flags |= FLAGS_KILLED;
            return exit_client(NULL, sptr, &me, "Nick change collision");
         }
         else
         {
            /*
             * same user detection
             * If they have the same user@host, we want to kill off the OLDER client to be nice
             * If they are different, kill off the NEWER client because it's trying to collide.
             *
             * They are the same user if... */
            sameuser = (mycmp(acptr->user->username, sptr->user->username) == 0) && /* they have the same user name */
                       (mycmp(acptr->user->host, sptr->user->host) == 0);           /* they have the same hostname */

            /*
             * If they're the same user, and the introduced has an older timestamp, OR
             * If they're NOT the same user, and the introduced has a newer timestamp
             */
            if ((sameuser && newts < acptr->tsinfo) || (!sameuser && newts > acptr->tsinfo))
            {
               /* this nickname (sptr) is undesireable. kill it. */
               if (sameuser)
                  sendto_ops_lev(SKILL_LEV, "Nick change override from %s to %s(%s <- %s)(older killed)",
                                 sptr->name, acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
               else
                  sendto_ops_lev(SKILL_LEV, "Nick change override from %s to %s(%s <- %s)(newer killed)",
                                 sptr->name, acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
               ircstp->is_kill++;
               /* 
                * tricky tricky!
                * us (or the servers behind us) have the right idea about this nick.
                * we want to kill off the new client being introduced by cptr
                * so _only_ send out this kill to cptr. -- before we sent to all
                * servers _but_ cptr - lucas
                *
                * Hrm. 
                * This seems like another place where we need to re-introduce
                * our NICK to cptr after we send the KILL. Possible desynch.
                *
                * Note that the NICK may already be in the sendq, and this KILL may simply
                * KILL off our new client. Could this cause some nasty loops? ie
                *  our NICK gets to remote server, and is rejected
                *  remote server sends KILL & NICK back to us
                *  our KILL gets to remote server and kills its user
                *  our second NICK gets to remote server and creates our user
                * ... I think TS should protect against that.
                */
               sendto_one_client_real(cptr, sptr, &me, &CMD_KILL, ":%s (Nick change override (%s(%s) -> %s(%s)))",
                                      me.name, acptr->from->name, acptr->name, get_client_name(cptr, FALSE), sptr->name);
               sptr->flags |= FLAGS_KILLED;
               if (sameuser)
                  return exit_client(cptr, sptr, &me, "Nick change override (old)");
               else
                  return exit_client(cptr, sptr, &me, "Nick change override (new)");
            }
            else
            {
               /* the nick we found (acptr) is undesireable. kill it. */
               if (sameuser)
                  sendto_ops_lev(SKILL_LEV, "Nick change collision from %s to %s(%s <- %s)(older killed)",
                                 sptr->name, acptr->name, acptr->from->name, get_client_name(cptr, FALSE));
               else
                  sendto_ops_lev(SKILL_LEV, "Nick change collision from %s to %s(%s <- %s)(newer killed)",
                                 sptr->name, acptr->name, acptr->from->name, get_client_name(cptr, FALSE));

               ircstp->is_kill++;
               send_me_numeric(acptr, ERR_NICKCOLLISION, acptr->name);
               send_serv_message_butone(cptr, &me, acptr, &CMD_KILL, ":%s (Nick change collide (%s(%s) <- %s(%s)))",
                                        me.name, acptr->from->name, acptr->name, get_client_name(cptr, FALSE), sptr->name);
               acptr->flags |= FLAGS_KILLED;
               exit_client(cptr, acptr, &me, "Nick collision");
               break; /* continue onward ... */
            }
         }
      }
   } while (0);

   /*
    * The nickname has passed all the collision checks.
    * Now, do the job of adding it or moving it around
    * on our lists, etc..
    */

   if (IsServer(sptr)) /* new NICK */
   {
      u_long newid, newip = 0;
      char *base64id = NULL;

      if(*parv[7] == '!' && parv[7][1])
      {
         char *ipp;
         aClient *bucptr;

         base64id = &parv[7][1];
         ipp = strchr(base64id, ':');
         if(ipp)
         {
            *ipp++ = '\0';
            newip = base64dec(ipp);
         }
		
         if((bucptr = find_by_base64_id(base64id)))
         {
            /* This should NEVER EVER EVER EVER happen!!!!! */
            sendto_realops("IDENTITY COLLISION! (%s[%s][%s] <> %s[%s][%s])", 
                           sptr->name, sptr->uplink->name, base64id,
                           bucptr->name, bucptr->uplink->name, bucptr->id.string);
            exit_client(cptr, bucptr, &me, "Identity collision!");
         }

         uplink = find_server_by_base64_id(base64id, &newid);
         /* kludge-ish for do_user() so we can stick name in scache */
         if(uplink)
            parv[7] = uplink->name;         
      }
      else
         uplink = find_server(parv[7], NULL);

      if(!uplink)
      {
         /* if we can't find the server this nick is on, complain loudly and ignore it. - lucas */
         sendto_realops("Remote nick %s on UNKNOWN server %s\n", nick, parv[7]);
         return 0;
      }
      sptr = make_client(cptr, uplink);

      /* If this is on a U: lined server, it's a U: lined client. */
      if(IsULine(uplink))
         sptr->flags |= FLAGS_ULINE;
                
      add_client_to_list(sptr);
      sptr->hopcount = atoi(parv[2]);
      if(newts)
         sptr->tsinfo = newts;
      else
      {
         newts = sptr->tsinfo = (ts_val) timeofday;
         ts_warn("Remote nick %s introduced without a TS", nick);
      }

      if(base64id)
      {
         SetHasID(sptr); /* this client is identity capable */
         sptr->id.id = newid; /* new identity! */
         strncpy(sptr->id.string, base64id, 8); /* save the base64 id so we don't have to keep generating it */
         add_userid_to_server(uplink, sptr); /* finally, add this user to the server hash record */
         sptr->ip.s_addr = newip; /* new IP address.. */
      }
      else if (parc == 11)
      {
	newip = strtoul(parv[9], NULL, 0);
	sptr->ip.s_addr = ntohl(newip);
      }

      strcpy(sptr->name, nick);
      add_to_client_hash_table(nick, sptr);

      /* adding the remote client to the throttle hash. 
       * checking local clients takes place in s_bsd.c
       * -epi */
      if (newip != 0)
	throttle_check(inetntoa((char *)&sptr->ip), -1, sptr->tsinfo);

      /* parse the usermodes intelligently */
      mptr = &parv[4][1];
      while (*mptr)
      {
         if((flag = user_mode_table[(unsigned char)*mptr]))
         {
            if ((flag == UMODE_o) || (flag == UMODE_O))
               Count.oper++;
            sptr->umode |= flag & SEND_UMODES;
         }
         mptr++;
      }                          
	  if (parc == 10)
	      return do_user(nick, cptr, sptr, parv[5], parv[6], parv[7], 
                     numbersare64 ? base64dec(parv[8]) : strtoul(parv[8], NULL, 0), 
		     parv[9]);
	  else 
	      return do_user(nick, cptr, sptr, parv[5], parv[6], parv[7],
			 numbersare64 ? base64dec(parv[8]) : strtoul(parv[8], NULL, 0),
			 parv[10]);
   }
   else /* remote nick change */
   {
      /*
       * Client just changing his/her nick. If he/she is on a
       * channel, send note of change to all clients on that channel.
       * Propagate notice to other servers.
       */
#ifndef DONT_CHECK_QLINE_REMOTE
      if ((aconf = find_conf_name(nick, CONF_QUARANTINED_NICK)))
      {
         sendto_realops("Q:lined nick %s from %s on %s", nick,
                        (*sptr->name != 0 && !IsServer(sptr)) ?
			sptr->name : "<unregistered>",
                        (sptr->user == NULL) ? ((IsServer(sptr)) ?
			parv[6] : me.name) : sptr->user->server);
      }
#endif
      /*
       * if the nickname is different, set the TS
       * AND set it -r. No need to propogate MODE -r and spam
       * the network on registered nick changes. yuck. - lucas
       */
      
       if (mycmp(parv[0], nick))
       {
          sptr->tsinfo = newts ? newts : (ts_val) timeofday;
          sptr->umode &= ~UMODE_r;
       }

       sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
       add_history(sptr, 1);
       send_serv_message_butone(cptr, sptr, NULL, &CMD_NICK, "%s %ld", nick, sptr->tsinfo);

       del_from_client_hash_table(sptr->name, sptr);
       samenick = mycmp(sptr->name, nick) ? 0 : 1;
       if (!samenick)
          hash_check_watch(sptr, RPL_LOGOFF);

       strcpy(sptr->name, nick);
       add_to_client_hash_table(nick, sptr);
       if (!samenick)
          hash_check_watch(sptr, RPL_LOGON);
       return 0;
   }

   /* we should never get here, ever ... */
   abort();

   return 0;
}

/* guaranteed sptr == cptr, !IsServer(cptr), !IsServer(sptr) */
int u_nick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aConfItem  *aconf;
   aClient    *acptr;
   chanMember *cm;
   char        nick[NICKLEN + 2];
   ts_val      newts = 0;
   int         samenick = 0;
        
   if (parc < 2) 
   {
      sendto_one(sptr, err_str(ERR_NONICKNAMEGIVEN), me.name, parv[0]);
      return 0;
   }
        
   parc = 2;
   /* parc == 2 on a normal client sign on (local) and a normal client nick change */

   strncpyzt(nick, parv[1], NICKLEN + 1);

   if (do_nick_name(nick) == -1) /* invalid nickname */
   {
      send_me_numeric(sptr, ERR_ERRONEUSNICKNAME, parv[1]);
      return 0;
   }

   do 
   {
#ifdef BROKEN_SERVER_NAMES
      if ((acptr = find_server(nick, NULL)))
      {
         send_me_numeric(sptr, ERR_NICKNAMEINUSE, nick);
         return 0;
      }
#endif /* BROKEN_SERVER_NAMES */

      /* nothing else is using this nick, great! */
      if (!(acptr = find_client(nick, NULL)))
        break;

      /* client changing nicks to itself.. */
      if (acptr == sptr)
      {
         if (strcmp(acptr->name, nick) != 0)
            break;
         else
            return 0;
      }

      if (IsUnknown(acptr))
      {
         exit_client(NULL, acptr, &me, "Overridden by older local signon");
         break;
      }

      /*
       * NICK is coming from local client connection. Just send
       * error reply and ignore the command.
       * parv[0] is empty on connecting clients
       */
      send_me_numeric(sptr, ERR_NICKNAMEINUSE, nick);
      return 0;
   } while (0);

   if (sptr->name[0])
   {
      /*
       * Client just changing his/her nick. If he/she is on a
       * channel, send note of change to all clients on that channel.
       * Propagate notice to other servers.
       */
      if ((aconf = find_conf_name(nick, CONF_QUARANTINED_NICK)) && !IsAnOper(sptr))
      {
         send_me_numeric(sptr, ERR_ERRONEUSNICKNAME, nick, BadPtr(aconf->passwd) ? 
                         "reason unspecified" : aconf->passwd);
         sendto_realops_lev(CCONN_LEV, "Forbidding Q:lined nick %s from %s.",
                        nick, get_client_name(cptr, FALSE));
         return 0;
      }

      /*
       * if the nickname is different, set the TS
       * AND set it -r. No need to propogate MODE -r and spam
       * the network on registered nick changes. yuck. - lucas
       */
      
       if (mycmp(parv[0], nick))
       {
          sptr->tsinfo = newts ? newts : (ts_val) timeofday;
          sptr->umode &= ~UMODE_r;
       }

       if (IsPerson(sptr))
       {

          /* before we change their nick, make sure they're not banned
           * on any channels, and!! make sure they're not changing to
           * a banned nick -sed */
          /* a little cleaner - lucas */

 	  for (cm = sptr->user->channel; cm; cm = cm->nextchan) 
          {
             if (can_send(sptr, cm->chptr)) 
             {
                send_me_numeric(sptr, ERR_BANNICKCHANGE, cm->chptr->chname);
                return 0;
             }
             if (nick_is_banned(cm->chptr, nick, sptr) != NULL) 
             {
                send_me_numeric(sptr, ERR_BANONCHAN, nick, cm->chptr->chname);
                return 0;
             }
          }

#ifdef ANTI_NICK_FLOOD
          if ((sptr->user->last_nick_change + MAX_NICK_TIME) < NOW)
             sptr->user->number_of_nick_changes = 0;
          sptr->user->last_nick_change = NOW;
          sptr->user->number_of_nick_changes++;
                                
          if (sptr->user->number_of_nick_changes <= MAX_NICK_CHANGES || IsAnOper(sptr))
          {
#endif
             sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
             if(IsPerson(sptr))
             {
                add_history(sptr, 1);
                send_serv_message_butone(cptr, sptr, NULL, &CMD_NICK, "%s %ld", nick, sptr->tsinfo);
             }
#ifdef ANTI_NICK_FLOOD
          }
          else
          {
             send_me_notice(sptr, ":*** Notice -- Too many nick changes. Wait %d seconds before trying again.", MAX_NICK_TIME);
             return 0;
          }
#endif
       }
       del_from_client_hash_table(sptr->name, sptr);
       samenick = mycmp(sptr->name, nick) ? 0 : 1;
       if (IsPerson(sptr) && !samenick)
         hash_check_watch(sptr, RPL_LOGOFF);
       strcpy(sptr->name, nick);
       add_to_client_hash_table(nick, sptr);
       if (IsPerson(sptr) && !samenick)
          hash_check_watch(sptr, RPL_LOGON);
       return 0;
   }
   else /* *sptr->name == '\0' */
   {
      /* Client setting NICK the first time */
      if ((aconf = find_conf_name(nick, CONF_QUARANTINED_NICK)))
      {
         send_me_numeric(sptr, ERR_ERRONEUSNICKNAME, nick,
                         BadPtr(aconf->passwd) ? "reason unspecified" : aconf->passwd);
         sendto_realops_lev(CCONN_LEV, "Forbidding Q:lined nick %s from <unregistered>%s.",
                        nick, get_client_name(cptr, FALSE));
         return 0;
      }

      strcpy(sptr->name, nick);
      sptr->tsinfo = timeofday;
      add_to_client_hash_table(nick, sptr);
      if (sptr->user)
      {
         /* USER already received, now we have NICK. */

          if (register_user(cptr, sptr, nick, sptr->user->username) == CLIENT_DIED)
             return CLIENT_DIED;
      }
      return 0;
   }
   /* we should never get here */
   abort();

   return 0;
}


