/* m_services.c - Because s_user.c was just crazy.
 *   Copyright (C) 1990 Jarkko Oikarinen and
 *                      University of Oulu, Computing Center
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free softwmare; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: m_services.c,v 1.6 2002/10/08 22:22:20 lucas Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include <sys/stat.h>
#include <utmp.h>
#include <fcntl.h>
#include "h.h"

/* Externally defined stuffs */
extern int user_modes[];
extern int user_mode_table[];


/*
 * the services aliases.
 *
 * NICKSERV  - /nickserv
 * CHANSERV  - /chanserv
 * OPERSERV  - /operserv
 * MEMOSERV  - /memoserv
 * SERVICES  - /services
 * IDENTIFY  - /identify
 */

int m_chanserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }
   if ((acptr = find_person_on_server(ChanServ, SERVICES_NAME)))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG_D, COLONPRE, parv[1]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, ChanServ);
   return 0;
}

int m_nickserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((acptr = find_person_on_server(NickServ, SERVICES_NAME)))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG_D, COLONPRE, parv[1]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, NickServ);
   return 0;
}

int m_memoserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((acptr = find_person_on_server(MemoServ, SERVICES_NAME)))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG_D, COLONPRE, parv[1]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, MemoServ);
   return 0;
}

int m_rootserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((acptr = find_person_on_server(RootServ, SERVICES_NAME)))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG_D, COLONPRE, parv
[1]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, RootServ);
   return 0;
}


int m_operserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((acptr = find_person_on_server(OperServ,
#ifdef OPERSERV_OTHER_HOST
       OPERSERV_OTHER_HOST
#else
       SERVICES_NAME
#endif
       )))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG, COLONPRE, parv[1]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, OperServ);
   return 0;
}

int m_statserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((acptr = find_person_on_server(StatServ,
#ifdef OPERSERV_OTHER_HOST
       OPERSERV_OTHER_HOST
#else
       SERVICES_NAME
#endif
       )))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG, COLONPRE, parv[1]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, StatServ);
   return 0;
}

int
m_helpserv(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
    aClient	*acptr;

    if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }
   if ((acptr = find_person_on_server(HelpServ,
#ifdef OPERSERV_OTHER_HOST
       OPERSERV_OTHER_HOST
#else
       SERVICES_NAME
#endif
       )))
      sendto_one_clientonserver_copy(acptr, sptr, &CMD_PRIVMSG, COLONPRE, parv[1
]);
   else
      send_me_numeric(sptr, ERR_SERVICESDOWN, HelpServ);
   return 0;
}


int m_services(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   char *tmps;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((!myncmp(parv[1], "help", 4)))
   {
      sendto_one(sptr, ":services!service@%s NOTICE %s :For ChanServ "
                 "help use: /chanserv help", SERVICES_NAME,
                 sptr->name);
      sendto_one(sptr, ":services!service@%s NOTICE %s :For NickServ "
                 "help use: /nickserv help", SERVICES_NAME,
                 sptr->name);
      sendto_one(sptr, ":services!service@%s NOTICE %s :For MemoServ "
                 "help use: /memoserv help", SERVICES_NAME,
                 sptr->name);
      return 0;
   }

   if ((tmps = (char *) strchr(parv[1], ' ')))
   {
      for(; *tmps == ' '; tmps++); /* find the first non-space after the command */

      if (*tmps == '#')
         return m_chanserv(cptr, sptr, parc, parv);
      else
         return m_nickserv(cptr, sptr, parc, parv);
   }
   return m_nickserv(cptr, sptr, parc, parv);
}

int m_identify(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient    *acptr;

   if (parc < 2 || *parv[1] == '\0')
   {
      send_me_numericNA(sptr, ERR_NOTEXTTOSEND);
      return 0;
   }

   if ((*parv[1] == '#') && ((char *) strchr(parv[1], ' ')))
   {
      if ((acptr = find_person_on_server(ChanServ, SERVICES_NAME)))
         sendto_one_clientonserver(acptr, sptr, &CMD_PRIVMSG_D, ":IDENTIFY %s", parv[1]);
      else
         send_me_numeric(sptr, ERR_SERVICESDOWN, ChanServ);
   }
   else
   {
      if ((acptr = find_person_on_server(NickServ, SERVICES_NAME)))
         sendto_one_clientonserver(acptr, sptr, &CMD_PRIVMSG_D, ":IDENTIFY %s", parv[1]);
      else
         send_me_numeric(sptr, ERR_SERVICESDOWN, NickServ);
  }
   return 0;
}


/* s_svsnick - Pretty straight forward.  Mostly straight outta df
 *  - Raistlin
 * parv[0] = sender
 * parv[1] = old nickname/id
 * parv[2] = new nickname
 * parv[3] = timestamp
 */
int m_svsnick(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aClient *acptr, *ocptr;

   if (!IsULine(sptr) || parc < 4 || (strlen(parv[2]) > NICKLEN))
      return 0;

   /* if we can't find them, just toss this message out... */
   if(!(ocptr = find_person_target(parv[1])) || !IsRegistered(ocptr))
      return 0;

   /* if we can't SVSNICK them to something because the nick is in use, KILL them */
   acptr = find_client(parv[2], NULL);
   if(acptr != NULL)
   {
      if(!IsRegistered(acptr))
         return exit_client(&me, acptr, &me, "SVSNICK Override");

      /* send a kill out for the person instead --wd */
      send_serv_message_butone(cptr, &me, ocptr, &CMD_KILL, ":%s (SVSNICK Collide [new])", sptr->name);
      /* now send back a kill for the nick they were presumably changed to */
      sendto_one_client_real(cptr, acptr, &me, &CMD_KILL, ":%s (SVSNICK Collide [old])", sptr->name);
      return 0;
   }

   if(use_or_deliver(cptr, sptr, &CMD_SVSNICK, "%s %s %s", 1, parc, parv) == HUNTED_ISME)
   {
      ocptr->umode &= ~UMODE_r;
      ocptr->tsinfo = strtol(parv[3], NULL, 0);
#ifdef ANTI_NICK_FLOOD
      ocptr->user->last_nick_change = strtol(parv[3], NULL, 0);
#endif
      sendto_common_channels(ocptr, ":%s NICK :%s", parv[1], parv[2]);
      add_history(ocptr, 1);
      send_serv_message_butone(NULL, ocptr, NULL, &CMD_NICK, "%s %ld", parv[2], ocptr->tsinfo);
      del_from_client_hash_table(ocptr->name, ocptr);
      strcpy(ocptr->name, parv[2]);
      add_to_client_hash_table(parv[2], ocptr);
   }
   return 0;
}

/* channel_svsmode:
 * parv[0] sender
 * parv[1] channel
 * parv[2] modes
 * parv[3] nick
 * parv[4] nickts
 * currently, only a mode of -b is supported.
 * services should use MODE for regular channel modes.
 * 2/5/00 lucas
 * preconditions: parc >= 3, sptr is ulined
 */
int channel_svsmode(aClient *cptr, aClient *sptr, int parc, char *parv[])
{
   aChannel *chptr;
   aClient *acptr = NULL;
   char *m, *nick = NULL;
   char change = '+';
   ts_val nickts = 0;
   int sendmsg = 1;

   if(!(chptr = find_channel(parv[1], NULL)))
      return 0;

   if(parc >= 4)
   {
      nick = parv[3];
      if(parc > 4)
         nickts = atol(parv[4]);
   }

   if(nick)
   {
      acptr = find_person(nick, NULL);
      if(!acptr || (nickts && acptr->tsinfo != nickts))
         return 0;
   }
   for(m = parv[2]; *m; m++)
      switch(*m)
      {
         case '+':
         case '-':
            change = *m;
            break;

         case 'b':
            if(nick && MyClient(acptr) && change == '-')
            {
               remove_matching_bans(chptr, acptr, &me);
               sendmsg--;
            }
            break;

         default:
            sendmsg++;
            break;
      }

   if(!sendmsg) return 0;
   if(nick)
      send_serv_message_butone(cptr, sptr, NULL, &CMD_SVSMODE, "%s %s %s %ld", parv[1], parv[2], nick, acptr->tsinfo);
   else
      send_serv_message_butone(cptr, sptr, NULL, &CMD_SVSMODE, "%s %s %s", parv[1], parv[2], nick);

   return 0;
}

/* m_svsmode - df function integrated
 *  - Raistlin
 * -- Behaviour changed - Epi (11/30/99)
 * parv[0] - sender
 * parv[1] - nick
 * parv[2] - TS (or mode, depending on svs version)
 * parv[3] - mode (or services id if old svs version)
 * parv[4] - optional argument (services id)
 */
int m_svsmode(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    int flag, what, oldumode;
    char *m, *modes, *optarg;
    aClient *acptr;
    ts_val ts = 0;

    if (!IsULine(sptr) || (parc < 3))
          return 0;

    if (parv[1][0] == '#')
       return channel_svsmode(cptr, sptr, parc, parv);

    if ((parc >= 4) && ((parv[3][0] == '+') || (parv[3][0] == '-')))
    {
      ts = atol(parv[2]);
      modes = parv[3];
      optarg = (parc > 4) ? parv[4] : NULL;
    }
    else
    {
      modes = parv[2];
      optarg = (parc > 3) ? parv[3] : NULL;
    }
    if (!(acptr = find_person_target(parv[1])))
      return 0;

    if (ts && (ts != acptr->tsinfo))
      return 0;

    what = MODE_ADD;
    oldumode = acptr->umode;

    for (m = modes; *m; m++)
       switch(*m) {
       case '+':
         what = MODE_ADD;
         break;
       case '-':
         what = MODE_DEL;
         break;
       case ' ':
       case '\n':
       case '\r':
       case '\t':
         break;
       case 'd':
         if (optarg && isdigit(*optarg))
            acptr->user->servicestamp = strtoul(optarg, NULL, 0);
         break;
       default:
         if((flag = user_mode_table[(unsigned char)*m]))
         {
            if (what == MODE_ADD)
               acptr->umode |= flag;
            else
               acptr->umode &= ~flag;
         }
         break;
    }

    if (optarg)
                send_serv_message_butone(cptr, sptr, acptr, &CMD_SVSMODE, "%ld %s %s",
                                         acptr->tsinfo, modes, optarg);
    else
                send_serv_message_butone(cptr, sptr, acptr, &CMD_SVSMODE, "%ld %s",
                                         acptr->tsinfo, modes);

    if (MyClient(acptr))
    {
        char buf[BUFSIZE];

        send_umode(acptr, acptr, oldumode, ALL_UMODES, buf);
    }

    return 0;
}


