/************************************************************************
 *   IRC - Internet Relay Chat, src/servconf.c
 *   Copyright (C) 2000 bahamut team
 *
 *   See file AUTHORS in IRC package for additional names of
 *   the programmers.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Id: servconf.c,v 1.5 2002/10/09 00:07:47 lucas Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "msg.h"
#include "channel.h"
#include "nameser.h"
#include "resolv.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <utmp.h>
#include "h.h"

/*
 * m_sqline (add an sqline)
 *
 * parv[1]  mask
 * parv[2]: reason
 */

int m_sqline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
   aConfItem *aconf;

   if(!(IsServer(sptr) || IsULine(sptr)))
      return 0;

   if(parc<2) 
   {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SQLINE");
	return 0;
   }
	
   /* get rid of redundancies */
   parv[1] = collapse(parv[1]);
   /* if we have any Q:lines (SQ or Q) that match
    * this Q:line, just return (no need to waste cpu */

   if (!(aconf=find_conf_name(parv[1], *parv[1]=='#' ? 
		CONF_QUARANTINED_CHAN : CONF_QUARANTINED_NICK)))
   {
	/* okay, it doesn't suck, build a new conf for it */
	aconf=make_conf();
        if(*parv[1]=='#')
            	aconf->status=(CONF_QUARANTINED_CHAN|CONF_SQLINE);
        else
		aconf->status=(CONF_QUARANTINED_NICK|CONF_SQLINE);
	DupString(aconf->name, parv[1]);
	DupString(aconf->passwd, (parv[2]!=NULL ? parv[2] : "Reserved"));
	aconf->next=conf;
	conf=aconf;
   }

   send_serv_message_butone(cptr, sptr, NULL, &CMD_SQLINE, "%s :%s", parv[1], aconf->passwd);
   return 0;
}

/*
 * m_unsqline
 * 
 * parv[1] mask or 0/1 (0 == mycmp, 1 == match)
 * parv[2] mask if parv[1] is 0/1
 */

int m_unsqline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
    aConfItem *aconf, *ac2, *ac3;
    int matchit = 0;
    char *mask;
    
    if(!(IsServer(sptr) || IsULine(sptr)))
	return 0;
    
    if(parc<2) 
    {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNSQLINE");
	return 0;
    }

    if(parc == 3)
    {
       matchit = atoi(parv[1]);
       mask = parv[2];
    }
    else
       mask = parv[1];

    /* find the sqline and remove it */
    /* this was way too complicated and ugly. Fixed. -lucas */
    /* Changed this to use match.  Seems to make more sense? */
    /* Therefore unsqline * will clear the sqline list. */
    
    ac2 = ac3 = aconf = conf;
    while(aconf)
    {
	if((aconf->status & (CONF_QUARANTINED_NICK))
	   && (aconf->status & (CONF_SQLINE)) && 
	   aconf->name && (matchit ? 
			   !match(mask, aconf->name) : 
			   !mycmp(mask, aconf->name))) 
	{
	    if (conf==aconf) 
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf=ac2;
	} else {
	    ac3=aconf;
	    aconf=aconf->next;
	}
    }

    if(parc == 3)
       send_serv_message_butone(cptr, sptr, NULL, &CMD_UNSQLINE, "%d :%s", matchit, mask);
    else
       send_serv_message_butone(cptr, sptr, NULL, &CMD_UNSQLINE, "%s", mask);

    return 0;
}

int m_sgline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
   aConfItem *aconf;
   int len;
   char *mask, *reason;
    
   if(!(IsServer(sptr) || IsULine(sptr)))
      return 0;

   if(parc<3) 
   {
	sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "SGLINE");
	return 0;
   }
	
   len = atoi(parv[1]);
   mask = parv[2];
   if ((strlen(mask) > len) && (mask[len])==':') 
   {
       mask[len] = '\0';
       reason = mask+len+1;
   } else { /* Bogus */
       return 0;
   }

   /* if we have any G:lines (SG or G) that match
    * this G:line, just return (no need to waste cpu */

   if (!(aconf=find_conf_name(mask, CONF_GLINE)))
   {
	/* okay, it doesn't suck, build a new conf for it */
	aconf=make_conf();
	
	aconf->status=(CONF_GLINE|CONF_SGLINE); /* G:line and SGline, woo */
	DupString(aconf->name, mask);
	DupString(aconf->passwd, reason ? reason : "Reason Unspecified");
	aconf->next=conf;
	conf=aconf;
   }

   send_serv_message_butone(cptr, sptr, NULL, &CMD_SGLINE, "%d :%s:%s", len,
                            aconf->name, aconf->passwd);
   return 0;
}
	
int m_unsgline(aClient *cptr, aClient *sptr, int parc, char *parv[]) 
{
   aConfItem *aconf, *ac2, *ac3;
   int matchit = 0;
   char *mask;   

   if(!(IsServer(sptr) || IsULine(sptr)))
      return 0;
   
   if(parc<2) 
   {
      sendto_one(sptr, err_str(ERR_NEEDMOREPARAMS), me.name, parv[0], "UNSGLINE");
      return 0;
   }

   if(parc == 3)
   {
      matchit = atoi(parv[1]);
      mask = parv[2];
   }
   else
      mask = parv[1];
   
   /* find the sgline and remove it */
   /* this was way too complicated and ugly. Fixed. -lucas */
   /* Changed this to use match.  Seems to make more sense? */
   /* Therefore unsgline * will clear the sgline list. */
    ac2 = ac3 = aconf = conf;
    while(aconf)
    {
	if((aconf->status & (CONF_GLINE))
	   && (aconf->status & (CONF_SGLINE)) && 
	   aconf->name && (matchit ?  
			   !match(mask, aconf->name) : 
			   !mycmp(mask, aconf->name))) 
	{
	    if (conf==aconf) 
		ac2 = ac3 = conf = aconf->next;
	    else
		ac2 = ac3->next = aconf->next;
	    MyFree(aconf->passwd);
	    MyFree(aconf->name);
	    MyFree(aconf);
	    aconf=ac2;
	} else {
	    ac3=aconf;
	    aconf=aconf->next;
	}
    }   

    if(parc == 3)
       send_serv_message_butone(cptr, sptr, NULL, &CMD_UNSGLINE, "%d :%s", matchit, mask);
    else
       send_serv_message_butone(cptr, sptr, NULL, &CMD_UNSGLINE, ":%s", mask);

   return 0;
}
