#!/bin/sh
spitshell=cat
package=IRC
md5sum=$1
md5sumflags=$2
sources=$3

echo "Extracting $package/src/version.c..."

creation=`date | \
awk '{if (NF == 6) \
{ print $1 " "  $2 " " $3 " "  $6 " at " $4 " " $5 } \
else \
{ print $1 " "  $2 " " $3 " " $7 " at " $4 " " $5 " " $6 }}'`

if test -r version.c.last
then
	generation=`sed -n 's/^char \*generation = \"\(.*\)\";/\1/p' < version.c.last`
	genesis=`sed -n 's/^char \*genesis = \"\(.*\)\";/\1/p' < version.c.last`
if test ! "$generation" ; then generation=0; fi
if test ! "$genesis" ; then genesis=$creation; fi
else
	generation=0
	genesis=$creation
fi

generation=`expr $generation + 1`

uname=`uname -a`
ulimit=`ulimit -a | sed 's/$/\",/;s/^/   "/'`

$spitshell >version.c <<!SUB!THIS!
/*
 *   IRC - Internet Relay Chat, src/version.c
 *   Copyright (C) 1990 Chelsea Ashley Dyerman
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * This file is generated by version.c.SH. Any changes made will go away.
 */

#include "struct.h"
#include "patchlevel.h"

char *generation = "$generation";
char *creation = "$creation";
char *genesis = "$genesis";
char version[128];

char *infotext[] =
    {
	 "$package --",
	 "Loosely based on the original code and concepts of Jarkko Oikarinen",
	 "Copyright 1988, 1989, 1990, 1991 University of Oulu, Computing Center",
	 "",
	 "This program is free software; you can redistribute it and/or",
	 "modify it under the terms of the GNU General Public License as",
	 "published by the Free Software Foundation; either version 1, or",
	 "(at your option) any later version.",
	 "",
	 "Send bug reports to dalnet-src@dal.net,",
	 "or security-related bug reports to coders@dal.net.",
	 "",
         "The bahamut team currently consists of the following people:",
	 "lucas            Lucas Madar   Coding Director   lucas@dal.net",
	 "Epiphani         Aaron Wiebe                     epiphani@dal.net",
	 "Raistlin_Majere  Jason Slagle                    raistlin@dal.net",
         "",
	 "The following people have contributed code, or support",
	 "to the bahamut team:",
	 "aClient             Diane Bruce         db@koruna.varner.com",
	 "Dakal               David Knepper       dakal@dal.net",
	 "Dalvenjah           Sven Nielsen        dalvenjah@dal.net",
	 "MSofty              Mark Salerno        msofty@dal.net",
	 "Rakarra             Ian Westcott        rakarra@dal.net",
	 "sedition            David Parton        sedition@dal.net",
	 "White_Dragon        Chip Norkus         wd@dal.net",
         "                    Michael Gold",
	 "",
	 "The following people have contributed ideas or testing to the",
	 "bahamut project:",
	 "driz                David Friedman      driz@dal.net",
	 "habit               Melinda Thompson    habit@dal.net",
	 "Kevin               Kevin Turner        kevin@dal.net",
	 "",
	 "",
	 "A special thanks to LMNOP for providing many laughs to the bahamut",
	 "team.",
	 "",
	 "",
	 "Developers and Contributors for previous versions:",
	 "Chris A. Bongaarts  Chris Behrens       Diane Bruce",
	 "Johannes Erdfelt    Roger Espel Llima   Jon Lusky",
	 "Michael Pearce      Brian Kraemer       Joan Touzet",
	 "Chris Portman       Keith Fralick       Sam Noble",
	 "Taner Halicioglu    Chris Behrens       Darren Reed",
	 "Markku Savela       Greg Lindahl        Jarkko Oikarinen",
	 "Armin Gruner        Matthew Green       Chuck Kane",
	 "Matt Lyle           Vesa Ruokonen       Nicolas PIOCH",
	 "Stellan Klebom      Dan Goodwin         Mike Bolotski",
	 "Ian Frechette       Markku Jarvinen     Kimmo Suominen",
	 "Jeff Trim           Vijay Subramaniam   Karl Kleinpaste",
	 "Bill Wisner         Tom Davis           Hugo Calendar",
	 "Tom Hopkins         Stephen van den Berg",
	 "Bo Adler            Michael Sandrof     Jon Solomon",
	 "Jan Peterson        Helen Rose          Paul Graham",
	 "",
	 "Thanks also goes to those persons not mentioned here who have added",
	 "their advice, opinions, and code to IRC.",
	 "Thanks also to those who provide the kind sys admins who let me and",
	 "others continue to develop IRC.",
	 "Thanks also to those who continue to make twinkies and doritos, as they",
	 "are the backbone of most programmers' diets and thus make it possible",
	 "for code like this to be developed!",
	 "",
	 0,
};

char *uname_result = "$uname";
char *ulimit_result[] = {
$ulimit
   0
};

char *checksums[] = {
!SUB!THIS!

echo "Extracting checksums..."

if [ ! -f $md5sum ]; then
   echo "Error: could not find md5sum utility in path '$md5sum'"
   exit 1
fi

for i in $sources version.c.SH .md5sums; do \
   echo "	\"`$md5sum $md5sumflags $i 2>/dev/null`\"," >> version.c; \
done

STATUS=`$md5sum $md5sumflags --check .md5sums 2>/dev/null`
iprev="dummy"
modded=""

for i in $STATUS; do
   if test $i = "FAILED"; then
      echo "	\"$iprev FAILED\", " >> version.c;
      modded="$modded `echo $iprev | sed s/://`";
   fi;
   iprev=$i;
done

echo "	0" >> version.c
echo "};" >> version.c

if test "$modded" != ""; then
   echo "**********************************************************************"
   echo "* Modified files detected"
   echo "**********************************************************************"
   for i in $modded; do
      if test $i = "s_err.c"; then
         echo "*** $i (modification does not need approval)"
      else
         echo "*** $i"
      fi
   done
   echo "**********************************************************************"
   echo "* Please make sure you have the permission of the coding team before"
   echo "* running custom patches on the main network."
   echo "**********************************************************************"
   sleep 1
fi
