/*  Internet Relay Chat (IRC) tools/convert.c 
 *
 *  Copyright (C) 2000 Aaron Wiebe (Epiphani@dal.net)
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

/* $Id: convert.c,v 1.5 2000/12/07 03:08:44 epiphani Exp $ */

#include "struct.h"
#include "common.h"
#include "sys.h"
#include "numeric.h"
#include "inet.h"
#include <sys/socket.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <signal.h>
#include "h.h"
#include "dich_conf.h"

#define MyMalloc(x) malloc(x)  /* lazy */

int   count;

typedef struct GLOBAL {
	char *bind;
	char *name;
	char *info;
	int   port;
	char *diepass;
	char *restartpass;
} GLOBAL;

typedef struct LINK {
	char *name;
	char *host;
	char *npass;
	char *cpass;
	int   port;
	int   hub;
        int   class;
	struct LINK *next;
} LINK;

int
dgets(int fd, char *buf, int num)
{
   static char dgbuf[8192];
   static char *head = dgbuf, *tail = dgbuf;
   char *s, *t;
   int n, nr;

   /*
    * * Sanity checks.
    */
   if (head == tail)
      *head = '\0';

   if (!num) {
      head = tail = dgbuf;
     *head = '\0';
      return 0;
   }

   if (num > sizeof(dgbuf) - 1)
      num = sizeof(dgbuf) - 1;

   FOREVER
   {
      if (head > dgbuf) {
         for (nr = tail - head, s = head, t = dgbuf; nr > 0; nr--)
            *t++ = *s++;
         tail = t;
         head = dgbuf;
      }
      /*
       * * check input buffer for EOL and if present return string.
       */
      if (head < tail &&
          ((s = strchr(head, '\n')) || (s = strchr(head, '\r'))) && s < tail) {
         n = MIN(s - head + 1, num);    /*
                                         * at least 1 byte
                                         */
         memcpy(buf, head, n);
         head += n;
         if (head == tail)
            head = tail = dgbuf;
         return n;
      }
      if (tail - head >= num) { /*
                                 * dgets buf is big enough
                                 */
         n = num;
         memcpy(buf, head, n);
         head += n;
         if (head == tail)
            head = tail = dgbuf;
         return n;
      }

      n = sizeof(dgbuf) - (tail - dgbuf) - 1;
      nr = read(fd, tail, n);
      if (nr == -1) {
         head = tail = dgbuf;
         return -1;
      }

      if (!nr) {
         if (tail > head) {
            n = MIN(tail - head, num);
            memcpy(buf, head, n);
            head += n;
            if (head == tail)
               head = tail = dgbuf;
            return n;
         }
         head = tail = dgbuf;
         return 0;
      }

      tail += nr;
      *tail = '\0';

      for (t = head; (s = strchr(t, '\n'));) {
         if ((s > head) && (s > dgbuf)) {
            t = s - 1;
            for (nr = 0; *t == '\\'; nr++)
               t--;
            if (nr & 1) {
               t = s + 1;
               s--;
               nr = tail - t;
               while (nr--)
                  *s++ = *t++;
               tail -= 2;
               *tail = '\0';
            }
            else
               s++;
         }
         else
            s++;
         t = s;
      }
      *tail = '\0';
   }
}

char *getfield(char *newline)
{
   static char *line = (char *) NULL;
   char       *end, *field;

   if (newline)
      line = newline;

   if (line == (char *) NULL)
      return ((char *) NULL);

   field = line;
   if ((end = strchr(line, ':')) == NULL) {
      line = (char *) NULL;
      if ((end = strchr(field, '\n')) == (char *) NULL)
         end = field + strlen(field);
   }
   else
      line = end + 1;
   *end = '\0';
   return (field);
}


int
main(int ARGC, char *ARGV[])
{

  char *filename, *tmp, *s, *t, *tmp2;
  char line[512], c[80];
  int   fd, i, p;
  GLOBAL *global;
  LINK   *link, *linkstart;


  static char quotes[9][2] =
  {
      {'b', '\b'},
      {'f', '\f'},
      {'n', '\n'},
      {'r', '\r'},
      {'t', '\t'},
      {'v', '\v'},
      {'\\', '\\'},
      {0, 0}
   };

  global = ((struct GLOBAL *) malloc(sizeof(struct GLOBAL)));
  link = ((struct LINK *) malloc(sizeof(struct LINK)));
  linkstart = link;

  if ((filename = ARGV[1]) == NULL)
  {
	printf("Useage: ./convert ircd.conf [ > ircd.bconf ]\n");
	exit(0);
  }

  printf("// Opening %s...", filename);

  if ((fd = ((int) open(filename, O_RDONLY))) <= NULL)
  {
        printf("FAILED\n");
        exit(-1);
   } else {
        printf("done.\n");
   }

   (void) dgets(-1, NULL, 0);
   while ((i = dgets(fd, line, sizeof(line) - 1)) > 0)
   {
      line[i] = '\0';
      if ((tmp = (char *) strchr(line, '\n')))
         *tmp = '\0';
      else
         while (dgets(fd, c, sizeof(c) - 1) > 0)
            if ((tmp = (char *) strchr(c, '\n')))
            {
               *tmp = '\0';
               break;
            }
      for (tmp = line; *tmp; tmp++)
      {
         if (*tmp == '\\')
         {
            for (i = 0; quotes[i][0]; i++)
               if (quotes[i][0] == *(tmp + 1))
               {
                  *tmp = quotes[i][1];
                  break;
               }
            if (!quotes[i][0])
               *tmp = *(tmp + 1);
            if (!*(tmp + 1))
               break;
            else
               for (s = tmp; (*s = *(s + 1)); s++);
         }
                  else if (*tmp == '#')
	{
	printf("// %s \n", tmp);
            *tmp = '\0';  
	}
      }
      if (!*line || line[0] == '#' || line[0] == '\n' ||
          line[0] == ' ' || line[0] == '\t')
         continue;
      if (line[1] != ':')
      {
      	   printf("// Bad configuration line: %s\n", line);
	   continue;
      }
      printf("// %s\n", line);

      t = getfield(line);
	switch(*t)
	{
         case 'A':
         case 'a':              /* Administrative info */
	    printf("admin {\n");
	    if ((tmp2 = getfield(NULL)) == NULL)
		continue;
	    else
		printf("\tlinea \"%s\";\n", tmp2);
	    if ((tmp2 = getfield(NULL)) == NULL)
		continue;
            else
                printf("\tlineb \"%s\";\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
		continue;
            else
                printf("\tlinec \"%s\";\n", tmp2);
            printf("};\n\n");
            break;

         case 'C':              /* Server where I should try to connect */
         case 'c':
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
	    {
		p = 0;
		if(link->name)
		{
			if (match(tmp2, link->host))
			{
		            if ((tmp2 = getfield(NULL)) == NULL)
		                continue;
            		    else
				DupString(link->cpass, tmp2);
			    tmp2 = getfield(NULL);
                            if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                            else
                                link->port = atoi(tmp2);
			    p = 1;
			 } else {
			    for (link = linkstart ; ((p == 1) || (link == NULL)) ;  link = link->next )
		            {
			   	if (match(tmp2, link->host))
			   	{
                              		if ((tmp2 = getfield(NULL)) == NULL)
                                		continue;
                              		else
                                		DupString(link->cpass, tmp2);
                              		tmp2 = getfield(NULL);
                              		if ((tmp2 = getfield(NULL)) == NULL)
                                 		continue;
                              		else
                                		link->port = atoi(tmp2);
					p = 1;
			        }
   			     }
			 }	
			 if (p == 1)
			   break;
			 else
			 {
				link->next = (struct LINK *) malloc(sizeof(struct LINK));
				link = link->next;
    	                        DupString(link->host, tmp2);
        	                if ((tmp2 = getfield(NULL)) == NULL)
                	                continue;
                       		else
                       		         DupString(link->cpass, tmp2);
                        	if ((tmp2 = getfield(NULL)) == NULL)
                                	continue;
   	                        else
        	                        DupString(link->name, tmp2);
                	        if ((tmp2 = getfield(NULL)) == NULL)
                        	        continue;
                	        else
                        	        link->port = atoi(tmp2);
                      		if ((tmp2 = getfield(NULL)) == NULL)
                                	continue;
                        	else
                                	link->class = atoi(tmp2);
			}
		} else {
			DupString(link->host, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(link->cpass, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(link->name, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                link->port = atoi(tmp2);
			if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                link->class = atoi(tmp2);
		}
	   }
           break;

         case 'Z':              /* Zap lines (immediate refusal)  */
         case 'z':
            printf("deny {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tip %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\treason \"%s\";\n", tmp2);
	    printf("};\n\n");
	    break;

         case 'E':              /* kline exception lines */
         case 'e':
	    printf("except {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\thost %s;\n", tmp2);
            tmp2 = getfield(NULL);
	    if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tident %s;\n", tmp2);
            printf("};\n\n");
            break;

         case 'F':              /* Super-Exempt hosts */
         case 'f':
	    printf("free {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\thost %s;\n", tmp2);
            tmp2 = getfield(NULL);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tident %s;\n", tmp2);
            printf("};\n\n");
            break;

         case 'G':              /* GECOS ban */
         case 'g':
	    printf("// GECOS bans not yet implemented in bindconf\n");
            break;

         case 'H':              /* Hub server line */
         case 'h':
            tmp2 = getfield(NULL);
            tmp2 = getfield(NULL);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
	    {
		p = 0;
		if(link->name)
		{
			if (match(tmp2, link->name))
			{
			    link->hub = 1;
			    p = 1;
			 } else {
			    for (link = linkstart ; ((p == 1) || (link == NULL)) ;  link = link->next )
		            {
			   	if (match(tmp2, link->name))
			   	{
					link->hub = 1;
					p = 1;
			        }
   			     }
			 }	
			 if (p == 1)
			   break;
			 else
			 {
				link->next = (struct LINK *) malloc(sizeof(struct LINK));
				link = link->next;
    	                        DupString(link->name, tmp2);
				link->hub = 1;
			}
		} else {
			DupString(link->name, tmp2);
			link->hub = 1;
		}
	   }
           break;

         case 'i':              /* to connect me */
         case 'I':
	    printf("allow {\n");	
	    if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\thost %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tpass %s;\n", tmp2);
            tmp2 = getfield(NULL);
            tmp2 = getfield(NULL);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tclass %s;\n", tmp2);
	    printf("};\n\n");
	    break;

         case 'K':              /* the infamous klines */
         case 'k':
	    printf("kline {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\thost %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\treason \"%s\";\n", tmp2);
            printf("};\n\n");

            break;

         case 'L':              /* guaranteed leaf server */
         case 'l':
            break;

         case 'M':
         case 'm':
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(global->name, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(global->bind, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(global->info, tmp2);
			if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                global->port = atoi(tmp2);

            break;

         case 'N':                  
         case 'n':              
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
	    {
		p = 0;
		if(link->name)
		{
			if (match(tmp2, link->host))
			{
		            if ((tmp2 = getfield(NULL)) == NULL)
		                continue;
            		    else
				DupString(link->npass, tmp2);
			    tmp2 = getfield(NULL);
                            if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                            else
                                link->port = atoi(tmp2);
			    p = 1;
			 } else {
			    for (link = linkstart ; ((p == 1) || (link == NULL)) ;  link = link->next )
		            {
			   	if (match(tmp2, link->host))
			   	{
                              		if ((tmp2 = getfield(NULL)) == NULL)
                                		continue;
                              		else
                                		DupString(link->npass, tmp2);
                              		tmp2 = getfield(NULL);
                              		if ((tmp2 = getfield(NULL)) == NULL)
                                 		continue;
                              		else
                                		link->port = atoi(tmp2);
					p = 1;
			        }
   			     }
			 }	
			 if (p == 1)
			   break;
			 else
			 {
				link->next = (struct LINK *) malloc(sizeof(struct LINK));
				link = link->next;
    	                        DupString(link->host, tmp2);
        	                if ((tmp2 = getfield(NULL)) == NULL)
                	                continue;
                       		else
                       		         DupString(link->npass, tmp2);
                        	if ((tmp2 = getfield(NULL)) == NULL)
                                	continue;
   	                        else
        	                        DupString(link->name, tmp2);
                	        if ((tmp2 = getfield(NULL)) == NULL)
                        	        continue;
                	        else
                        	        link->port = atoi(tmp2);
                      		if ((tmp2 = getfield(NULL)) == NULL)
                                	continue;
                        	else
                                	link->class = atoi(tmp2);
			}
		} else {
			DupString(link->host, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(link->npass, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                DupString(link->name, tmp2);
                        if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                link->port = atoi(tmp2);
			if ((tmp2 = getfield(NULL)) == NULL)
                                continue;
                        else
                                link->class = atoi(tmp2);
		}
	   }
           break;


         case 'O':              /* Operator line */
         case 'o':
	    printf("oper {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\thost %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tpass %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tnick %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tflags %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tclass %s;\n", tmp2);
	    printf("};\n\n");
            break;

         case 'P':              /* listen port line */
         case 'p':
	    printf("bind {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tallow %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\taddress %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tpass %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tport %s;\n", tmp2);
	    printf("};\n\n");
            break;

         case 'Q':              /* restricted nicknames */
	 case 'q':
	    printf("qline {\n");
            tmp2 = getfield(NULL);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\treason \"%s\";\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tnick %s;\n", tmp2);
	    printf("};\n\n");	
            break;

         case 'T':
         case 't':
	    printf("wgmon {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\thost %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\turl %s;\n", tmp2);
	    printf("};\n\n");
	    break;

         case 'U':              /* Ultimate Servers (aka God) */
         case 'u':
	    printf("super {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tname %s;\n", tmp2);
	    printf("};\n\n");
            break;

         case 'X':              /* die/restart pass line */
         case 'x':
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
		DupString(global->diepass, tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
		DupString(global->restartpass, tmp2);
            break;

         case 'Y':              /* Class line */
         case 'y':
	    printf("class {\n");
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tname %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tping %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tconnect %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tmax %s;\n", tmp2);
            if ((tmp2 = getfield(NULL)) == NULL)
                continue;
            else
                printf("\tsendq %s;\n", tmp2);
	    printf("};\n\n");
            break;

         default:
            printf("// Error in config file: %s\n\n", line);
            break;
     	}
}
link = linkstart;
for (link; link ; link = link->next)
{	printf("link {\n");
	printf("\tname %s;\n", link->name);
	printf("\thost %s;\n", link->host);
	printf("\tcpass %s;\n", link->cpass);
	printf("\tnpass %s;\n", link->npass);
	printf("\tident %d;\n", link->port);
	printf("\tclass %d;\n", link->class);
	if(link->hub == 1)
		printf("\thub 1;\n");
	printf("};\n\n");
}
printf("global {\n");
printf("\tname %s;\n", global->name);
printf("\tbind %s;\n", global->bind);
printf("\tinfo \"%s\";\n", global->info);
printf("\tport %d;\n", global->port);
if(global->diepass)
	printf("\tdiepass %s;\n", global->diepass);
if(global->restartpass)
	printf("\trestartpass %s;\n", global->restartpass);
printf("};\n\n");
return 1;
}

/************************************************************************
 *   IRC - Internet Relay Chat, src/match.c
 *   Copyright (C) 1990 Jarkko Oikarinen
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 1, or (at your option)
 *   any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#if 0 /* this code is depricated, however it may still prove useful
         in the long run to 'show' it here */

#define MAX_CALLS 512
static int  calls = 0;

/*
 * *  Compare if a given string (name) matches the given *  mask (which
 * can contain wild cards: '*' - match any *  number of chars, '?' -
 * match any single character. *
 * 
 *      return  0, if match *           1, if no match
 */
/*
 * * _match() * Iterative matching function, rather than recursive. *
 * Written by Douglas A Lewis (dalewis@acsu.buffalo.edu)
 */

static int wdmatch(const char *string, const char *wild);

static int _match(char *mask, char *name)
{
   u_char *m = (u_char *) mask, *n = (u_char *) name;
   char       *ma = mask, *na = name;
   int         wild = 0, q = 0;

   while (1) {
      if (calls++ > MAX_CALLS)
	 return 1;
      if (*m == '*') {
	 while (*m == '*')
	    m++;
	 wild = 1;
	 ma = (char *) m;
	 na = (char *) n;
      }

      if (!*m) {
	 if (!*n)
	    return 0;
	 for (m--; (m > (u_char *) mask) && (*m == '?'); m--);
	 if ((*m == '*') && (m > (u_char *) mask) &&
	     (m[-1] != '\\'))
	    return 0;
	 if (!wild)
	    return 1;
	 m = (u_char *) ma;
	 n = (u_char *) ++na;
      }
      else if (!*n) {
	 while (*m == '*')
	    m++;
	 return (*m != 0);
      }
      if ((*m == '\\') && ((m[1] == '*') || (m[1] == '?'))) {
	 m++;
	 q = 1;
      }
      else
	 q = 0;

      if ((tolower(*m) != tolower(*n)) && ((*m != '?') || q)) {
	 if (!wild)
	    return 1;
	 m = (u_char *) ma;
	 n = (u_char *) ++na;
      }
      else {
	 if (*m)
	    m++;
	 if (*n)
	    n++;
      }
   }
}

int match(char *n, char *m) {
	calls=0;
	return wdmatch(n,m);
}
#endif

/*
 * * collapse a pattern string into minimal components. * This
 * particular version is "in place", so that it changes the pattern *
 * which is to be reduced to a "minimal" size.
 */
char       *
collapse(char *pattern)
{
   char   *s = pattern, *s1, *t;

   if (BadPtr(pattern))
      return pattern;
   /*
    * Collapse all \** into \*, \*[?]+\** into \*[?]+
    */
   for (; *s; s++)
      if (*s == '\\') {
	 if (!*(s + 1))
	    break;
	 else
	    s++;
      } else if (*s == '*') {
	 if (*(t = s1 = s + 1) == '*')
	    while (*t == '*')
	       t++;
	 else if (*t == '?')
	    for (t++, s1++; *t == '*' || *t == '?'; t++)
	       if (*t == '?')
		  *s1++ = *t;
	 while ((*s1++ = *t++));
      }
   return pattern;
}

unsigned char tolowertab[] =
{0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
 0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
 0x1e, 0x1f,
 ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
 '*', '+', ',', '-', '.', '/',
 '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
 ':', ';', '<', '=', '>', '?',
 '@', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
 't', 'u', 'v', 'w', 'x', 'y', 'z', '[', '\\', ']', '^',
 '_',
 '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i',
 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's',
 't', 'u', 'v', 'w', 'x', 'y', 'z', '{', '|', '}', '~',
 0x7f,
 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff};

unsigned char touppertab[] =
{0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7, 0x8, 0x9, 0xa,
 0xb, 0xc, 0xd, 0xe, 0xf, 0x10, 0x11, 0x12, 0x13, 0x14,
 0x15, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d,
 0x1e, 0x1f,
 ' ', '!', '"', '#', '$', '%', '&', 0x27, '(', ')',
 '*', '+', ',', '-', '.', '/',
 '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
 ':', ';', '<', '=', '>', '?',
 '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^',
 0x5f,
 '`', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S',
 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '{', '|', '}', '~',
 0x7f,
 0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89,
 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
 0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99,
 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
 0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9,
 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
 0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9,
 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
 0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9,
 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
 0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9,
 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
 0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7, 0xe8, 0xe9,
 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
 0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9,
 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff};

unsigned char char_atribs[] =
{
    /*
     * 0-7 
     */ CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL,
    /*
     * 8-12 
     */ CNTRL, CNTRL | SPACE, CNTRL | SPACE, CNTRL | SPACE, CNTRL | SPACE,
    /*
     * 13-15 
     */ CNTRL | SPACE, CNTRL, CNTRL,
    /*
     * 16-23 
     */ CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL,
    /*
     * 24-31 
     */ CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL, CNTRL,
    /*
     * space 
     */ PRINT | SPACE,
    /*
     * !"#$%&'( 
     */ PRINT, PRINT, PRINT, PRINT, PRINT, PRINT, PRINT, PRINT,
    /*
     * )*+,-./ 
     */ PRINT, PRINT, PRINT, PRINT, PRINT, PRINT, PRINT,
    /*
     * 0123 
     */ PRINT | DIGIT, PRINT | DIGIT, PRINT | DIGIT, PRINT | DIGIT,
    /*
     * 4567 
     */ PRINT | DIGIT, PRINT | DIGIT, PRINT | DIGIT, PRINT | DIGIT,
    /*
     * 89:; 
     */ PRINT | DIGIT, PRINT | DIGIT, PRINT, PRINT,
    /*
     * <=>? 
     */ PRINT, PRINT, PRINT, PRINT,
    /*
     * @ 
     */ PRINT,
    /*
     * ABC 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * DEF 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * GHI 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * JKL 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * MNO 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * PQR 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * STU 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * VWX 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * YZ[ 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * \]^ 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * _   
     */ PRINT,
    /*
     * abc 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * def 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * ghi 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * jkl 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * mno 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * pqr 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * stu 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * vwx 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * yz{ 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * \}~ 
     */ PRINT | ALPHA, PRINT | ALPHA, PRINT | ALPHA,
    /*
     * del 
     */ 0,
    /*
     * 80-8f 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * 90-9f 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * a0-af 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * b0-bf 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * c0-cf 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * d0-df 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * e0-ef 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    /*
     * f0-ff 
     */ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* screw that lexocographic crap, nothing in ircd ever cares */
int
mycmp(char *s1, char *s2)
{
   unsigned char *str1 = (unsigned char *) s1;
   unsigned char *str2 = (unsigned char *) s2;
	while (touppertab[*str1] == touppertab[*str2]) {
		if (*str1 == '\0')
		  return 0;
		str1++;str2++;
	}
   return (-1);
}

int
myncmp(char *str1, char *str2, int n)
{
   unsigned char *s1 = (unsigned char *) str1;
   unsigned char *s2 = (unsigned char *) str2;
   while (touppertab[*s1] == touppertab[*s2]) {
		s1++;
		s2++;
      n--;
      if (n == 0 || (*s1 == '\0' && *s2 == '\0'))
		  return 0;
   }
   return (-1);
}
 
int match(char *wild, char *string) {
   /* the *! in a match is such a common case that we optimize
	 * for it inherently */
	if(wild[0]=='*' && wild[1]=='!') {
		wild+=2;
		while(*string!='!' && *string)
		  string++;
		if(!*string)
		  return 1;
		string++;
	}
	/* we also optimize further for *!*@...which is also very
	 * common...*/
	if(wild[0]=='*' && wild[1]=='@') {
		wild+=2;
		while(*string!='@' && *string)
		  string++;
		if(!*string)
		  return 1;
		string++;
	}
	while(1) {
		/*
		 * Logic: Match * in a string, this is confusing, sort of
		 * if * is the last thing in the wildcard, the match is
		 * definite if we've gotten this far.
		 * otherwise we try and find every occurance of the
		 * the next character in the wildcard within the string
		 * and match from there, calling the function recursively
		 * until some level below us returns positive, in which
		 * case we too return positive.  For strings with lots
		 * of wildcards this gets disgustingly recursive.
		 */
		if(!*wild)
		  return ((!*string) ? 0 : 1);
		if(*wild=='*') {
			/* if ret==2 then we know it's a problem un-solvable
			 * by match, so we totally give up instead of trying
			 * a bunch of other recursions to check through */
			wild++;
			/* swallow all extraneous '*'s after.  */
			while(*wild=='*')
			  wild++;
			while(*wild=='?' && *string) {
				wild++;string++;
			}
			if(!*wild)
			  return 0;
			if(*wild=='*')
			  continue;
			while(*string) {
				if(touppertab[(u_char)*string]==touppertab[(u_char)*wild] &&
					!match((wild+1), (string+1)))
				  return 0;
				string++;
			}
		}
		if(!*string)
		  return 1;
		if(*wild!='?' && touppertab[(u_char)*string]!=touppertab[(u_char)*wild])
			return 1;
		string++;
		wild++;
	}
	return 1;
}
